@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
  *
- * Copyright (C) 2009 Richard Hughes <richard@hughsie.com>
+ * Copyright (C) 2009-2010 Richard Hughes <richard@hughsie.com>
  *
  * Licensed under the GNU General Public License Version 2
  *
@@ -24,7 +24,7 @@
  * @short_description: Client object to hold an array of devices.
  *
  * This object holds an array of %GcmDevices, and watches both udev and xorg
- * for changes. If a device is added or removed then a signal is fired.
+ * for changes. If a device is added, removed or changed then a signal is fired.
  */
 
 #include "config.h"
@@ -34,6 +34,7 @@
 #include <gudev/gudev.h>
 #include <libgnomeui/gnome-rr.h>
 #include <cups/cups.h>
+#include <sane/sane.h>
 
 #include "gcm-client.h"
 #include "gcm-device-xrandr.h"
@@ -51,6 +52,8 @@ static void     gcm_client_finalize	(GObject     *object);
 #define GCM_CLIENT_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GCM_TYPE_CLIENT, GcmClientPrivate))
 
 static void gcm_client_xrandr_add (GcmClient *client, GnomeRROutput *output);
+static gboolean gcm_client_add_connected_devices_sane (GcmClient *client, GError **error);
+static gpointer gcm_client_add_connected_devices_sane_thrd (GcmClient *client);
 
 /**
  * GcmClientPrivate:
@@ -80,13 +83,17 @@ enum {
 enum {
 	SIGNAL_ADDED,
 	SIGNAL_REMOVED,
+	SIGNAL_CHANGED,
 	SIGNAL_LAST
 };
 
 static guint signals[SIGNAL_LAST] = { 0 };
+static gpointer gcm_client_object = NULL;
 
 G_DEFINE_TYPE (GcmClient, gcm_client, G_TYPE_OBJECT)
 
+#define	GCM_CLIENT_SANE_REMOVED_TIMEOUT		200	/* ms */
+
 /**
  * gcm_client_set_use_threads:
  **/
@@ -140,6 +147,22 @@ gcm_client_get_devices (GcmClient *client)
 	return g_ptr_array_ref (client->priv->array);
 }
 
+/**
+ * gcm_client_get_loading:
+ *
+ * @client: a valid %GcmClient instance
+ *
+ * Gets the loading status.
+ *
+ * Return value: %TRUE if the object is still loading devices
+ **/
+gboolean
+gcm_client_get_loading (GcmClient *client)
+{
+	g_return_val_if_fail (GCM_IS_CLIENT (client), FALSE);
+	return client->priv->loading;
+}
+
 /**
  * gcm_client_get_device_by_id:
  *
@@ -175,6 +198,17 @@ gcm_client_get_device_by_id (GcmClient *client, const gchar *id)
 	return device;
 }
 
+/**
+ * gcm_client_device_changed_cb:
+ **/
+static void
+gcm_client_device_changed_cb (GcmDevice *device, GcmClient *client)
+{
+	/* emit a signal */
+	egg_debug ("emit changed: %s", gcm_device_get_id (device));
+	g_signal_emit (client, signals[SIGNAL_CHANGED], 0, device);
+}
+
 /**
  * gcm_client_get_id_for_udev_device:
  **/
@@ -196,47 +230,44 @@ gcm_client_get_id_for_udev_device (GUdevDevice *udev_device)
 /**
  * gcm_client_gudev_remove:
  **/
-static void
+static gboolean
 gcm_client_gudev_remove (GcmClient *client, GUdevDevice *udev_device)
 {
 	GcmDevice *device;
 	gchar *id;
-	gboolean ret;
-	GcmClientPrivate *priv = client->priv;
+	gboolean ret = FALSE;
 
 	/* generate id */
 	id = gcm_client_get_id_for_udev_device (udev_device);
 
 	/* do we have a device that matches */
 	device = gcm_client_get_device_by_id (client, id);
-	if (device == NULL)
-		goto out;
-
-	/* remove device first as we hold a reference to it */
-	ret = g_ptr_array_remove (priv->array, device);
-	if (!ret) {
-		egg_warning ("failed to remove %s", id);
+	if (device == NULL) {
+		egg_debug ("no match for %s", id);
 		goto out;
 	}
 
-	/* emit a signal */
-	egg_debug ("emit removed: %s", id);
-	g_signal_emit (client, signals[SIGNAL_REMOVED], 0, device);
+	/* set disconnected */
+	gcm_device_set_connected (device, FALSE);
+
+	/* success */
+	ret = TRUE;
 out:
 	if (device != NULL)
 		g_object_unref (device);
 	g_free (id);
+	return ret;
 }
 
 /**
  * gcm_client_gudev_add:
  **/
-static void
+static gboolean
 gcm_client_gudev_add (GcmClient *client, GUdevDevice *udev_device)
 {
 	GcmDevice *device = NULL;
 	GcmDevice *device_tmp = NULL;
-	gboolean ret;
+	gboolean ret = FALSE;
 	const gchar *value;
 	GError *error = NULL;
 	GcmClientPrivate *priv = client->priv;
@@ -258,13 +289,10 @@ gcm_client_gudev_add (GcmClient *client, GUdevDevice *udev_device)
 	/* we might have a previous saved device with this ID, in which case nuke it */
 	device_tmp = gcm_client_get_device_by_id (client, gcm_device_get_id (device));
 	if (device_tmp != NULL) {
-
-		/* remove from the array */
-		g_ptr_array_remove (client->priv->array, device_tmp);
-
-		/* emit a signal */
-		egg_debug ("emit removed: %s", gcm_device_get_id (device));
-		g_signal_emit (client, signals[SIGNAL_REMOVED], 0, device_tmp);
+		/* already disconnected device now connected */
+		g_object_unref (device);
+		device = g_object_ref (device_tmp);
+		gcm_device_set_connected (device, TRUE);
 	}
 
 	/* load the device */
@@ -281,11 +309,50 @@ gcm_client_gudev_add (GcmClient *client, GUdevDevice *udev_device)
 	/* signal the addition */
 	egg_debug ("emit: added %s to device list", gcm_device_get_id (device));
 	g_signal_emit (client, signals[SIGNAL_ADDED], 0, device);
+
+	/* connect to the changed signal */
+	g_signal_connect (device, "changed", G_CALLBACK (gcm_client_device_changed_cb), client);
 out:
 	if (device != NULL)
 		g_object_unref (device);
 	if (device_tmp != NULL)
 		g_object_unref (device_tmp);
+	return ret;
+}
+
+/**
+ * gcm_client_sane_refresh_cb:
+ **/
+static gboolean
+gcm_client_sane_refresh_cb (GcmClient *client)
+{
+	gboolean ret;
+	GError *error = NULL;
+	GThread *thread;
+
+	/* inform UI if we are loading devices still */
+	client->priv->loading_refcount = 1;
+	gcm_client_set_loading (client, TRUE);
+
+	/* rescan */
+	egg_debug ("rescanning sane");
+	if (client->priv->use_threads) {
+		thread = g_thread_create ((GThreadFunc) gcm_client_add_connected_devices_sane_thrd, client, FALSE, &error);
+		if (thread == NULL) {
+			egg_debug ("failed to rescan sane devices: %s", error->message);
+			g_error_free (error);
+			goto out;
+		}
+	} else {
+		ret = gcm_client_add_connected_devices_sane (client, &error);
+		if (!ret) {
+			egg_debug ("failed to rescan sane devices: %s", error->message);
+			g_error_free (error);
+			goto out;
+		}
+	}
+out:
+	return FALSE;
 }
 
 /**
@@ -294,10 +361,42 @@ out:
 static void
 gcm_client_uevent_cb (GUdevClient *gudev_client, const gchar *action, GUdevDevice *udev_device, GcmClient *client)
 {
-	if (g_strcmp0 (action, "remove") == 0)
-		gcm_client_gudev_remove (client, udev_device);
-	else if (g_strcmp0 (action, "add") == 0)
-		gcm_client_gudev_add (client, udev_device);
+	gboolean ret;
+	const gchar *value;
+	GcmDevice *device_tmp;
+	guint i;
+	GcmClientPrivate *priv = client->priv;
+
+	if (g_strcmp0 (action, "remove") == 0) {
+		ret = gcm_client_gudev_remove (client, udev_device);
+		if (ret)
+			egg_debug ("removed %s", g_udev_device_get_sysfs_path (udev_device));
+
+		/* we need to remove scanner devices */
+		value = g_udev_device_get_property (udev_device, "GCM_RESCAN");
+		if (g_strcmp0 (value, "scanner") == 0) {
+
+			/* set all scanners as disconnected */
+			for (i=0; i<priv->array->len; i++) {
+				device_tmp = g_ptr_array_index (priv->array, i);
+				if (gcm_device_get_kind (device_tmp) == GCM_DEVICE_KIND_SCANNER)
+					gcm_device_set_connected (device_tmp, FALSE);
+			}
+
+			/* find any others that might still be connected */
+			g_timeout_add (GCM_CLIENT_SANE_REMOVED_TIMEOUT, (GSourceFunc) gcm_client_sane_refresh_cb, client);
+		}
+
+	} else if (g_strcmp0 (action, "add") == 0) {
+		ret = gcm_client_gudev_add (client, udev_device);
+		if (ret)
+			egg_debug ("added %s", g_udev_device_get_sysfs_path (udev_device));
+
+		/* we need to rescan scanner devices */
+		value = g_udev_device_get_property (udev_device, "GCM_RESCAN");
+		if (g_strcmp0 (value, "scanner") == 0)
+			g_timeout_add (GCM_CLIENT_SANE_REMOVED_TIMEOUT, (GSourceFunc) gcm_client_sane_refresh_cb, client);
+	}
 }
 
 /**
@@ -330,6 +429,7 @@ gcm_client_add_connected_devices_udev (GcmClient *client, GError **error)
 
 	/* inform the UI */
 	gcm_client_done_loading (client);
+
 	return TRUE;
 }
 
@@ -401,7 +501,6 @@ gcm_client_get_device_by_window (GcmClient *client, GdkWindow *window)
 	GnomeRROutput *output_best = NULL;
 	GnomeRROutput **outputs;
 	GcmDevice *device = NULL;
-	gchar *id = NULL;
 
 	/* get the window parameters, in root co-ordinates */
 	gdk_window_get_origin (window, &window_x, &window_y);
@@ -458,9 +557,8 @@ out:
 		device_tmp = gcm_device_xrandr_new ();
 		gcm_device_xrandr_set_from_output (device_tmp, output_best, NULL);
 		device = gcm_client_get_device_by_id (client, gcm_device_get_id (device_tmp));
-		g_free (device_tmp);
+		g_object_unref (device_tmp);
 	}
-	g_free (id);
 	return device;
 }
 
@@ -493,27 +591,20 @@ gcm_client_xrandr_add (GcmClient *client, GnomeRROutput *output)
 		goto out;
 	}
 
-	/* we might have a previous saved device with this ID, in which case nuke it */
+	/* we might have a previous saved device with this ID, in which use that instead */
 	device_tmp = gcm_client_get_device_by_id (client, gcm_device_get_id (device));
 	if (device_tmp != NULL) {
 
-		/* old device not connected */
-		g_object_get (device_tmp,
-			      "connected", &connected,
-			      NULL);
-
 		/* this is just a dupe */
+		connected = gcm_device_get_connected (device_tmp);
 		if (connected) {
 			egg_debug ("ignoring dupe");
 			goto out;
 		}
 
-		/* remove from the array */
-		g_ptr_array_remove (client->priv->array, device_tmp);
-
-		/* emit a signal */
-		egg_debug ("emit removed: %s (unconnected device that has just been connected)", gcm_device_get_id (device_tmp));
-		g_signal_emit (client, signals[SIGNAL_REMOVED], 0, device_tmp);
+		/* use original device */
+		g_object_unref (device);
+		device = g_object_ref (device_tmp);
 	}
 
 	/* load the device */
@@ -530,6 +621,9 @@ gcm_client_xrandr_add (GcmClient *client, GnomeRROutput *output)
 	/* signal the addition */
 	egg_debug ("emit: added %s to device list", gcm_device_get_id (device));
 	g_signal_emit (client, signals[SIGNAL_ADDED], 0, device);
+
+	/* connect to the changed signal */
+	g_signal_connect (device, "changed", G_CALLBACK (gcm_client_device_changed_cb), client);
 out:
 	if (device != NULL)
 		g_object_unref (device);
@@ -589,6 +683,9 @@ gcm_client_cups_add (GcmClient *client, cups_dest_t dest)
 	/* signal the addition */
 	egg_debug ("emit: added %s to device list", gcm_device_get_id (device));
 	g_signal_emit (client, signals[SIGNAL_ADDED], 0, device);
+
+	/* connect to the changed signal */
+	g_signal_connect (device, "changed", G_CALLBACK (gcm_client_device_changed_cb), client);
 out:
 	if (device != NULL)
 		g_object_unref (device);
@@ -628,6 +725,105 @@ gcm_client_add_connected_devices_cups_thrd (GcmClient *client)
 	return NULL;
 }
 
+/**
+ * gcm_client_sane_add:
+ **/
+static void
+gcm_client_sane_add (GcmClient *client, const SANE_Device *sane_device)
+{
+	gboolean ret;
+	GError *error = NULL;
+	GcmDevice *device = NULL;
+	GcmDevice *device_tmp = NULL;
+	GcmClientPrivate *priv = client->priv;
+
+	/* create new device */
+	device = gcm_device_sane_new ();
+	ret = gcm_device_sane_set_from_device (device, sane_device, &error);
+	if (!ret) {
+		egg_debug ("failed to set for output: %s", error->message);
+		g_error_free (error);
+		goto out;
+	}
+
+	/* we might have a previous saved device with this ID, in which use that instead */
+	device_tmp = gcm_client_get_device_by_id (client, gcm_device_get_id (device));
+	if (device_tmp != NULL) {
+		gcm_device_set_connected (device_tmp, TRUE);
+		g_object_unref (device);
+		device = g_object_ref (device_tmp);
+	}
+
+	/* load the device */
+	ret = gcm_device_load (device, &error);
+	if (!ret) {
+		egg_warning ("failed to load: %s", error->message);
+		g_error_free (error);
+		goto out;
+	}
+
+	/* add to the array */
+	g_ptr_array_add (priv->array, g_object_ref (device));
+
+	/* signal the addition */
+	egg_debug ("emit: added %s to device list", gcm_device_get_id (device));
+	g_signal_emit (client, signals[SIGNAL_ADDED], 0, device);
+
+	/* connect to the changed signal */
+	g_signal_connect (device, "changed", G_CALLBACK (gcm_client_device_changed_cb), client);
+out:
+	if (device != NULL)
+		g_object_unref (device);
+	if (device_tmp != NULL)
+		g_object_unref (device_tmp);
+}
+
+/**
+ * gcm_client_add_connected_devices_sane:
+ **/
+static gboolean
+gcm_client_add_connected_devices_sane (GcmClient *client, GError **error)
+{
+	gint i;
+	SANE_Status status;
+	const SANE_Device **device_list;
+
+	/* force sane to drop it's cache of devices -- yes, it is that crap */
+	sane_exit ();
+	sane_init (NULL, NULL);
+
+	/* get scanners on the local server */
+	status = sane_get_devices (&device_list, FALSE);
+	if (status != SANE_STATUS_GOOD) {
+		egg_warning ("failed to get devices from SANE: %s", sane_strstatus (status));
+		goto out;
+	}
+
+	/* nothing */
+	if (device_list == NULL || device_list[0] == NULL) {
+		egg_debug ("no devices to add");
+		goto out;
+	}
+
+	/* add them */
+	for (i=0; device_list[i] != NULL; i++)
+		gcm_client_sane_add (client, device_list[i]);
+out:
+	/* inform the UI */
+	gcm_client_done_loading (client);
+	return TRUE;
+}
+
+/**
+ * gcm_client_add_connected_devices_sane_thrd:
+ **/
+static gpointer
+gcm_client_add_connected_devices_sane_thrd (GcmClient *client)
+{
+	gcm_client_add_connected_devices_sane (client, NULL);
+	return NULL;
+}
+
 /**
  * gcm_client_add_unconnected_device:
  **/
@@ -635,10 +831,10 @@ static void
 gcm_client_add_unconnected_device (GcmClient *client, GKeyFile *keyfile, const gchar *id)
 {
 	gchar *title;
-	gchar *type_text = NULL;
+	gchar *kind_text = NULL;
 	gchar *colorspace_text = NULL;
-	GcmColorspaceEnum colorspace;
-	GcmDeviceTypeEnum type;
+	GcmColorspace colorspace;
+	GcmDeviceKind kind;
 	GcmDevice *device = NULL;
 	gboolean ret;
 	gboolean virtual;
@@ -650,40 +846,40 @@ gcm_client_add_unconnected_device (GcmClient *client, GKeyFile *keyfile, const g
 	if (title == NULL)
 		goto out;
 	virtual = g_key_file_get_boolean (keyfile, id, "virtual", NULL);
-	type_text = g_key_file_get_string (keyfile, id, "type", NULL);
-	type = gcm_device_type_enum_from_string (type_text);
-	if (type == GCM_DEVICE_TYPE_ENUM_UNKNOWN)
+	kind_text = g_key_file_get_string (keyfile, id, "type", NULL);
+	kind = gcm_device_kind_from_string (kind_text);
+	if (kind == GCM_DEVICE_KIND_UNKNOWN)
 		goto out;
 
 	/* get colorspace */
 	colorspace_text = g_key_file_get_string (keyfile, id, "colorspace", NULL);
 	if (colorspace_text == NULL) {
 		egg_warning ("legacy device %s, falling back to RGB", id);
-		colorspace = GCM_COLORSPACE_ENUM_RGB;
+		colorspace = GCM_COLORSPACE_RGB;
 	} else {
-		colorspace = gcm_colorspace_enum_from_string (colorspace_text);
+		colorspace = gcm_colorspace_from_string (colorspace_text);
 	}
 
-	/* create device or specified type */
+	/* create device of specified type */
 	if (virtual) {
 		device = gcm_device_virtual_new ();
-	} else if (type == GCM_DEVICE_TYPE_ENUM_DISPLAY) {
+	} else if (kind == GCM_DEVICE_KIND_DISPLAY) {
 		device = gcm_device_xrandr_new ();
-	} else if (type == GCM_DEVICE_TYPE_ENUM_PRINTER) {
+	} else if (kind == GCM_DEVICE_KIND_PRINTER) {
 		device = gcm_device_cups_new ();
-	} else if (type == GCM_DEVICE_TYPE_ENUM_CAMERA) {
+	} else if (kind == GCM_DEVICE_KIND_CAMERA) {
 		/* FIXME: use GPhoto? */
 		device = gcm_device_udev_new ();
-	} else if (type == GCM_DEVICE_TYPE_ENUM_SCANNER) {
+	} else if (kind == GCM_DEVICE_KIND_SCANNER) {
 		device = gcm_device_sane_new ();
 	} else {
-		egg_warning ("device type internal error");
+		egg_warning ("device kind internal error");
 		goto out;
 	}
 
 	/* create device */
 	g_object_set (device,
-		      "type", type,
+		      "kind", kind,
 		      "id", id,
 		      "connected", FALSE,
 		      "title", title,
@@ -706,10 +902,13 @@ gcm_client_add_unconnected_device (GcmClient *client, GKeyFile *keyfile, const g
 	/* signal the addition */
 	egg_debug ("emit: added %s to device list", id);
 	g_signal_emit (client, signals[SIGNAL_ADDED], 0, device);
+
+	/* connect to the changed signal */
+	g_signal_connect (device, "changed", G_CALLBACK (gcm_client_device_changed_cb), client);
 out:
 	if (device != NULL)
 		g_object_unref (device);
-	g_free (type_text);
+	g_free (kind_text);
 	g_free (colorspace_text);
 	g_free (title);
 }
@@ -753,9 +952,7 @@ gcm_client_add_saved (GcmClient *client, GError **error)
 			gcm_client_add_unconnected_device (client, keyfile, groups[i]);
 		} else {
 			egg_debug ("found already added %s", groups[i]);
-			g_object_set (device,
-				      "saved", TRUE,
-				      NULL);
+			gcm_device_set_saved (device, TRUE);
 		}
 	}
 out:
@@ -781,8 +978,8 @@ gcm_client_add_connected (GcmClient *client, GError **error)
 	if (!ret)
 		goto out;
 
-	/* inform UI if we are loading devces still */
-	client->priv->loading_refcount = 2;
+	/* inform UI if we are loading devices still */
+	client->priv->loading_refcount = 3;
 	gcm_client_set_loading (client, TRUE);
 
 	/* UDEV */
@@ -806,18 +1003,29 @@ gcm_client_add_connected (GcmClient *client, GError **error)
 		if (!ret)
 			goto out;
 	}
+
+	/* SANE */
+	if (client->priv->use_threads) {
+		thread = g_thread_create ((GThreadFunc) gcm_client_add_connected_devices_sane_thrd, client, FALSE, error);
+		if (thread == NULL)
+			goto out;
+	} else {
+		ret = gcm_client_add_connected_devices_sane (client, error);
+		if (!ret)
+			goto out;
+	}
 out:
 	return ret;
 }
 
 /**
- * gcm_client_add_device:
+ * gcm_client_add_virtual_device:
  **/
 gboolean
-gcm_client_add_device (GcmClient *client, GcmDevice *device, GError **error)
+gcm_client_add_virtual_device (GcmClient *client, GcmDevice *device, GError **error)
 {
 	gboolean ret = FALSE;
-	gchar *id = NULL;
+	const gchar *id;
 	gboolean virtual;
 	GcmDevice *device_tmp = NULL;
 
@@ -825,16 +1033,14 @@ gcm_client_add_device (GcmClient *client, GcmDevice *device, GError **error)
 	g_return_val_if_fail (GCM_IS_DEVICE (device), FALSE);
 
 	/* check removable */
-	g_object_get (device,
-		      "virtual", &virtual,
-		      "id", &id,
-		      NULL);
+	virtual = gcm_device_get_virtual (device);
 	if (!virtual) {
 		g_set_error_literal (error, 1, 0, "cannot add non-virtual devices");
 		goto out;
 	}
 
 	/* look to see if device already exists */
+	id = gcm_device_get_id (device);
 	device_tmp = gcm_client_get_device_by_id (client, id);
 	if (device_tmp != NULL) {
 		/* TRANSLATORS: error message */
@@ -846,18 +1052,18 @@ gcm_client_add_device (GcmClient *client, GcmDevice *device, GError **error)
 	g_ptr_array_add (client->priv->array, g_object_ref (device));
 
 	/* update status */
-	g_object_set (device,
-		      "saved", FALSE,
-		      NULL);
+	gcm_device_set_saved (device, FALSE);
 
 	/* emit a signal */
 	egg_debug ("emit added: %s", id);
 	g_signal_emit (client, signals[SIGNAL_ADDED], 0, device);
 
+	/* connect to the changed signal */
+	g_signal_connect (device, "changed", G_CALLBACK (gcm_client_device_changed_cb), client);
+
 	/* all okay */
 	ret = TRUE;
 out:
-	g_free (id);
 	if (device_tmp != NULL)
 		g_object_unref (device_tmp);
 	return ret;
@@ -870,20 +1076,18 @@ gboolean
 gcm_client_delete_device (GcmClient *client, GcmDevice *device, GError **error)
 {
 	gboolean ret = FALSE;
-	gchar *id = NULL;
+	const gchar *id;
 	gchar *data = NULL;
 	gchar *filename = NULL;
 	GKeyFile *keyfile = NULL;
+	gboolean saved;
 	gboolean connected;
 
 	g_return_val_if_fail (GCM_IS_CLIENT (client), FALSE);
 	g_return_val_if_fail (GCM_IS_DEVICE (device), FALSE);
 
-	/* check removable */
-	g_object_get (device,
-		      "connected", &connected,
-		      "id", &id,
-		      NULL);
+	/* check device is not connected */
+	connected = gcm_device_get_connected (device);
 	if (connected) {
 		g_set_error_literal (error, 1, 0, "device is still connected");
 		goto out;
@@ -896,6 +1100,12 @@ gcm_client_delete_device (GcmClient *client, GcmDevice *device, GError **error)
 		goto out;
 	}
 
+	/* check device is saved */
+	id = gcm_device_get_id (device);
+	saved = gcm_device_get_saved (device);
+	if (!saved)
+		goto not_saved;
+
 	/* get the config file */
 	filename = gcm_utils_get_default_config_location ();
 	egg_debug ("removing %s from %s", id, filename);
@@ -924,15 +1134,13 @@ gcm_client_delete_device (GcmClient *client, GcmDevice *device, GError **error)
 		goto out;
 
 	/* update status */
-	g_object_set (device,
-		      "saved", FALSE,
-		      NULL);
+	gcm_device_set_saved (device, FALSE);
 
+not_saved:
 	/* emit a signal */
 	egg_debug ("emit removed: %s", id);
 	g_signal_emit (client, signals[SIGNAL_REMOVED], 0, device);
 out:
-	g_free (id);
 	g_free (data);
 	g_free (filename);
 	if (keyfile != NULL)
@@ -1061,6 +1269,16 @@ gcm_client_class_init (GcmClientClass *klass)
 			      NULL, NULL, g_cclosure_marshal_VOID__OBJECT,
 			      G_TYPE_NONE, 1, G_TYPE_OBJECT);
 
+	/**
+	 * GcmClient::changed
+	 **/
+	signals[SIGNAL_CHANGED] =
+		g_signal_new ("changed",
+			      G_TYPE_FROM_CLASS (object_class), G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (GcmClientClass, changed),
+			      NULL, NULL, g_cclosure_marshal_VOID__OBJECT,
+			      G_TYPE_NONE, 1, G_TYPE_OBJECT);
+
 	g_type_class_add_private (klass, sizeof (GcmClientPrivate));
 }
 
@@ -1071,6 +1289,7 @@ static void
 gcm_client_init (GcmClient *client)
 {
 	const gchar *subsystems[] = {"usb", "video4linux", NULL};
+	SANE_Status status;
 
 	client->priv = GCM_CLIENT_GET_PRIVATE (client);
 	client->priv->display_name = NULL;
@@ -1089,6 +1308,11 @@ gcm_client_init (GcmClient *client)
 	/* for CUPS */
 	httpInitialize();
 
+	/* for SANE */
+	status = sane_init (NULL, NULL);
+	if (status != SANE_STATUS_GOOD)
+		egg_warning ("failed to init SANE: %s", sane_strstatus (status));
+
 	/* should be okay for localhost */
 	client->priv->http = httpConnectEncrypt (cupsServer (), ippPort (), cupsEncryption ());
 }
@@ -1101,12 +1325,21 @@ gcm_client_finalize (GObject *object)
 {
 	GcmClient *client = GCM_CLIENT (object);
 	GcmClientPrivate *priv = client->priv;
+	GcmDevice *device;
+	guint i;
+
+	/* do not respond to changed events */
+	for (i=0; i<priv->array->len; i++) {
+		device = g_ptr_array_index (priv->array, i);
+		g_signal_handlers_disconnect_by_func (device, G_CALLBACK (gcm_client_device_changed_cb), client);
+	}
 
 	g_free (priv->display_name);
 	g_ptr_array_unref (priv->array);
 	g_object_unref (priv->gudev_client);
 	g_object_unref (priv->screen);
 	httpClose (priv->http);
+	sane_exit ();
 
 	G_OBJECT_CLASS (gcm_client_parent_class)->finalize (object);
 }
@@ -1119,8 +1352,12 @@ gcm_client_finalize (GObject *object)
 GcmClient *
 gcm_client_new (void)
 {
-	GcmClient *client;
-	client = g_object_new (GCM_TYPE_CLIENT, NULL);
-	return GCM_CLIENT (client);
+	if (gcm_client_object != NULL) {
+		g_object_ref (gcm_client_object);
+	} else {
+		gcm_client_object = g_object_new (GCM_TYPE_CLIENT, NULL);
+		g_object_add_weak_pointer (gcm_client_object, &gcm_client_object);
+	}
+	return GCM_CLIENT (gcm_client_object);
 }
 
